import itertools
import numpy as np
import spglib as spg
from collections import Counter
from ase.neighborlist import NeighborList

from .atoms import Atom, Atoms
from .clusters import get_clusters
from .orbits import get_orbits
from ..io.logging import logger
from .eigentensors import create_eigentensors as _create_eigentensors
from .tensors import rotate_tensor_precalc, rotation_tensor_as_matrix
from .translational_constraints import create_constraint_map as _create_constraint_map  # noqa
from .utilities import BiMap


# TODO: Add longer description of each step
# TODO: Be careful with side effects
# TODO: Preferably the functions should not take the cs as input
def build_cluster_space(cluster_space, prototype_structure):

    """ The permutation list is an indexed fast lookup table for permutation
    vectors.
    """
    logger.debug('Populate permutation list')
    _create_permutations(cluster_space)

    """ The primitive cell is calcualted by spglib and contains the cell
    metric, basis and atomic numbers. After this the prototype structure is
    disposed.
    """
    logger.debug('Get primitive cell')
    _create_primitive_cell(cluster_space, prototype_structure)

    """ The symmetries are calculated by spglib and the main information is the
    rotation matrices and translation vectors for each symmetry in scaled
    coordinates.
    """
    logger.debug('Get symmetries')
    _create_symmetry_dataset(cluster_space)

    """ The neigbor atoms to the center cell are stored in an indexed list. The
    list contains all atoms within the maximum cutoff specified.
    """
    logger.debug('Find neighbors')
    _create_atom_list(cluster_space)

    """ Clusters are generated as combinations of the indices of the atoms in
    the atom_list. If the cluster is to be included or not depends on the
    specification of valid clusters in the cutoffs object. Often all atoms must
    be within some distance from each other which may depend on both expansion
    order and number of atoms in the cluster
    """
    logger.debug('Starting generating clusters')
    _create_cluster_list(cluster_space)
    logger.debug('Finished generating clusters')

    """ The clusters are categorized into orbits and the eigensymmetries are
    stored in each orbit.
    """
    logger.debug('Starting categorizing clusters')
    _create_orbits(cluster_space)
    logger.debug('Finished categorizing clusters')

    """ The eigensymmetries from the previous step is used to generate valid
    eigentensors
    """
    logger.debug('Starting finding eigentensors')
    _create_eigentensors(cluster_space)
    logger.debug('Finished finding eigentensors')

    """ If some orbits can't have a force constant due to symmetry they are
    dropped from the _orbits attribute but kept in a separate list named
    _dropped_orbits
    """
    logger.debug('Dropping orbits...')
    _drop_orbits(cluster_space)

    """ Each orientation family gets populated with its rotated version of the
    orbits eigentensors.
    """
    logger.debug('Rotating eigentensors into ofs...')
    _populate_ofs_with_ets(cluster_space)

    """ The matrix describing the mapping which preserves the global symmetries
    is created (translational and rotational symmetry).
    """
    logger.debug('Constructing constraint map')
    _create_constraint_map(cluster_space)
    logger.info('Constraints:')
    logger.info('    Acoustic: {}'.format(cluster_space.acoustic_sum_rules))
    ndofs_by_order = {o: cluster_space.get_n_dofs_by_order(o) for o in
                      cluster_space.cutoffs.orders}
    logger.info('    Number of degrees of freedom: {}'
                .format(ndofs_by_order))
    logger.info('    Total number of degrees of freedom: {}'
                .format(cluster_space.n_dofs))

    """  The eigentensors are rescaled depending on order to create a more well
    behaved system for fitting. NOTE! In here the constratints are rescaled too
    """
    logger.debug('Rescale eigentensors')
    _rescale_eigentensors(cluster_space)

    logger.debug('Normalize constraints')
    _normalize_constraint_vectors(cluster_space)

    logger.debug('Rotate tensors to Carteesian coordinates')
    _rotate_eigentensors(cluster_space)


# TODO: Actual input could be just the maximum order
# TODO: No side effects, returns only the permutation BiMap
def _create_permutations(cs):
    orders = cs.cutoffs.orders

    permutations = BiMap()
    for order in orders:
        for permutation in itertools.permutations(range(order)):
            permutations.append(permutation)

    cs._permutations = permutations


# TODO: tolerances must be fixed in a coherent way. Prefarably via a config
# object
# TODO: Add kw prim to CS init. To use the supplied primitive cell instead of
# having spglib return a funky rotated cell
# TODO: Does the basis check need to be done? There might not be a problem that
# it is close to 1 instead of 0 anymore. If thats the case it is better to keep
# it as spglib returns it
# TODO: Add good debug
# TODO: Assert spos dot cell == pos. Sometimes the positions can be outside of
# the cell and then there is a mismatch between wath is returned by
# get_sclad_positions, positions and what spos dot cell gives (it should give
# the position
# TODO: Check basis to see if it can be represented by sympy. (in preparation
# for rotational sum rules)
# TODO: general function -> break out into utility function
# TODO: Send the tolerance as a parameter instef of the whole cs.
def _create_primitive_cell(cs, prototype_structure):

    spgPrim = spg.standardize_cell(prototype_structure, no_idealize=True,
                                   to_primitive=True, symprec=cs.symprec)

    numbers_match = sorted(spgPrim[2]) == sorted(prototype_structure.numbers)
    spg_cell_volume = np.abs(np.linalg.det(spgPrim[0].T))
    prototype_cell_volume = np.abs(np.linalg.det(prototype_structure.cell.T))
    # TODO: is symprec the best tolerance to use for volume check?
    cell_volume_match = np.isclose(spg_cell_volume, prototype_cell_volume,
                                   atol=cs.symprec, rtol=0)

    if numbers_match and cell_volume_match:
        prim = Atoms(prototype_structure)
        prim.wrap()
    else:
        basis = spgPrim[1]
        if np.any(spgPrim[1] > (1 - cs.symprec)):
            logger.debug('Found basis close to 1:\n' +
                         ' {}'.format(str(spgPrim[1])))
            basis = spgPrim[1].round(8) % 1  # TODO
            logger.debug('Wrapping to:\n' +
                         ' {}'.format(str(basis)))

        prim = Atoms(cell=spgPrim[0],
                     scaled_positions=basis,
                     numbers=spgPrim[2], pbc=True)

    # log primitive cell information
    logger.info('Primitive cell:')
    logger.info('    Formula: {}'.format(prim.get_chemical_formula()))
    logger.info(('    Cell:' + '\n        [{:9.5f} {:9.5f} {:9.5f}]'*3).format(
        *prim.cell[0], *prim.cell[1], *prim.cell[2]))
    logger.info('    Basis:')
    if len(prim) < 5:
        for symbol, spos in zip(prim.get_chemical_symbols(), prim.basis):
            logger.info('        {:2}  [{:9.5f} {:9.5f} {:9.5f}]'.format(
                symbol, *spos))
    else:
        for sym, spos in zip(prim[:3].get_chemical_symbols(), prim[:3].basis):
            logger.info('        {:2}  [{:9.5f} {:9.5f} {:9.5f}]'.format(
                sym, *spos))
        logger.info('        ...')
    logger.info('')
    cs._prim = prim


# TODO: Fix how the tolerance is handled
# TODO: Look over properties to acccess symmetry_dataset. Especially rotation,
# translation and wyckoff
# TODO: Sen prim and symprec as parameters
def _create_symmetry_dataset(cs):

    prim = cs._prim

    symmetry_dataset = spg.get_symmetry_dataset(prim, symprec=cs.symprec)

    cs._symmetry_dataset = symmetry_dataset

    logger.info('Crystal symmetry:')
    logger.info('    Spacegroup:          {}'.format(cs.spacegroup))
    logger.info('    Unique atoms:        {}'.format(
        len(set(cs.wyckoff_sites))))
    logger.info('    Symmetry operations: {}'.format(
        len(cs.rotation_matrices)))
    logger.info('    symprec:             {:.2e}'.format(cs.symprec))
    logger.info('')


# TODO: Fix how the tolerance is handled
# TODO: Refactor the two runs of finding the neighbors
# TODO: The bug that the cutoff is exactly on a shell might be a non issue.
# TODO: It is possible to check that the clusters map out the orbits completely
# TODO: Send in prim, cutoff and config and return atom_list instead
def _create_atom_list(cs):

    tol = cs.symprec

    atom_list = BiMap()

    # Populating the atom list with the center atoms
    for i in range(len(cs._prim)):
        atom_list.append(Atom(i, [0, 0, 0]))

    logger.info('Cutoffs:')
    logger.info('    Maximum cutoff: {}'.format(cs.cutoffs.max_cutoff))
    # Find all the atoms which is neighbors to the atoms in the center cell
    # The pair cutoff should be larger or equal than the others
    cutoffs = [(cs.cutoffs.max_cutoff - tol) / 2] * len(cs._prim)
    nl = NeighborList(cutoffs=cutoffs, skin=0, self_interaction=True,
                      bothways=True)
    nl.update(cs._prim)
    for i in range(len(cs._prim)):
        for index, offset in zip(*nl.get_neighbors(i)):
            atom = Atom(index, offset)
            if atom not in atom_list:
                atom_list.append(atom)

    nl = NeighborList(
        cutoffs=[(cs.cutoffs.max_cutoff + tol) / 2] *
        len(cs._prim), skin=0, self_interaction=True,
        bothways=True)
    nl.update(cs._prim)
    distance_from_cutoff = tol
    for i in range(len(cs._prim)):
        for index, offset in zip(*nl.get_neighbors(i)):
            atom = Atom(index, offset)
            # ... and check that no new atom is found
            if atom not in atom_list:
                pos = atom.pos(cs._prim.basis, cs._prim.cell)
                distance = min(np.linalg.norm(pos - atom.position)
                               for atom in cs._prim) - cs.cutoffs.max_cutoff
                distance_from_cutoff = min(distance, distance_from_cutoff)

    if distance_from_cutoff != tol:
        msg = 'Maximum cutoff close to neighbor shell, change cutoff'
        raise Exception(msg)

    msg = '    Found {} center atom{} with {} images totaling {} atoms'.format(
        len(cs._prim), 's' if len(cs._prim) > 1 else '',
        len(atom_list) - len(cs._prim), len(atom_list))
    logger.info(msg)
    logger.info('')

    cs._atom_list = atom_list


# TODO: add atoms property to cs
# TODO: Only inputs are prim, atom_list and cutoffs
def _create_cluster_list(cs):

    # Convert the atom list from site/offset to scaled positions
    spos = [a.spos(cs._prim.basis) for a in cs._atom_list]
    numbers = [cs._prim.numbers[a.site] for a in cs._atom_list]
    # Make an atoms object out of the scaled positions
    atoms = Atoms(cell=cs._prim.cell,
                  scaled_positions=spos,
                  numbers=numbers,
                  pbc=False)

    cs._cluster_filter.setup(atoms)

    cs._cluster_list = get_clusters(atoms, cs.cutoffs, len(cs._prim))

    logger.info('Clusters:')
    counter = Counter(len(c) for c in cs._cluster_list)
    logger.info('    Clusters: {}'.format(dict(counter)))
    logger.info('    Total number of clusters: {}\n'
                .format(sum(counter.values())))


def _create_orbits(cs):
    # TODO: Check scaled/cart
    cs._orbits = get_orbits(cs._cluster_list,
                            cs._atom_list,
                            cs.rotation_matrices,
                            cs.translation_vectors,
                            cs.permutations,
                            cs._prim,
                            cs.symprec)
    orbits_to_drop = []
    for i, orbit in enumerate(cs.orbits):
        if not cs._cluster_filter(cs._cluster_list[orbit.prototype_index]):
            orbits_to_drop.append(i)

    reduced_orbits = []
    cs._dropped_orbits = []

    for i in range(len(cs.orbits)):
        if i in orbits_to_drop:
            cs._dropped_orbits.append(cs.orbits[i])
        else:
            reduced_orbits.append(cs.orbits[i])
    cs._orbits = reduced_orbits

    counter = Counter(orb.order for orb in cs._orbits)
    logger.info('Orbits:')
    logger.info('    Orbits: {}'.format(dict(counter)))
    logger.info('    Total number of orbits: {}\n'
                .format(sum(counter.values())))


def _drop_orbits(cs):
    orbits_to_drop = []
    for i, orbit in enumerate(cs.orbits):
        if not orbit.eigentensors:
            orbits_to_drop.append(i)

    reduced_orbits = []

    for i in range(len(cs.orbits)):
        if i in orbits_to_drop:
            cs._dropped_orbits.append(cs.orbits[i])
        else:
            reduced_orbits.append(cs.orbits[i])
    cs._orbits = reduced_orbits

    logger.info('Eigentensors:')
    n_ets = dict()
    for order in cs.cutoffs.orders:
        n_ets[order] = sum(len(orb.eigentensors) for orb in cs.orbits
                           if orb.order == order)
    logger.info('    Eigentensors: {}'.format(n_ets))
    logger.info('    Total number of parameters: {}'
                .format(sum(n_ets.values())))
    if len(cs._dropped_orbits) > 0:
        logger.info('    Discarded orbits:')
        for orb in cs._dropped_orbits:
            logger.info('        {}'
                        .format(cs.cluster_list[orb.prototype_index]))
    logger.info('')


def _populate_ofs_with_ets(cs):

    R_inv_lookup = dict()
    for orbit_index, orbit in enumerate(cs.orbits):
        for of in orbit.orientation_families:
            R_inv_lookup_index = (of.symmetry_index, orbit.order)
            R_inv = R_inv_lookup.get(R_inv_lookup_index, None)
            if R_inv is None:
                R = cs.rotation_matrices[of.symmetry_index]
                R_inv_tmp = np.linalg.inv(R)
                R_inv = R_inv_tmp.astype(np.int64)
                assert np.allclose(R_inv, R_inv_tmp), (R_inv, R_inv_tmp)
                R_inv = rotation_tensor_as_matrix(R_inv, orbit.order)
                R_inv_lookup[R_inv_lookup_index] = R_inv
            of.eigentensors = []
            for et in orbit.eigentensors:
                rotated_et = rotate_tensor_precalc(et, R_inv)
                assert rotated_et.dtype == np.int64
                of.eigentensors.append(rotated_et)


def _rotate_eigentensors(cs):
    V_invT = np.linalg.inv(cs._prim.cell.T)
    lookup = dict()
    for orb in cs.orbits:
        V_invT_tensormatrix = lookup.get(orb.order, None)
        if V_invT_tensormatrix is None:
            V_invT_tensormatrix = rotation_tensor_as_matrix(V_invT, orb.order)
            lookup[orb.order] = V_invT_tensormatrix
        orb.eigentensors = [rotate_tensor_precalc(et, V_invT_tensormatrix) for et in orb.eigentensors]  # noqa
        for of in orb.orientation_families:
            of.eigentensors = [rotate_tensor_precalc(et, V_invT_tensormatrix) for et in of.eigentensors]  # noqa


def _normalize_constraint_vectors(cs):
    M = cs._cvs
    norms = np.zeros(M.shape[1])
    for c, v in zip(M.col, M.data):
        norms[c] += v**2
    for i in range(len(norms)):
        norms[i] = np.sqrt(norms[i])
    for i, c in enumerate(M.col):
        M.data[i] /= norms[c]


def _rescale_eigentensors(cs):
    for orbit in cs.orbits:
        norm = cs.length_scale**orbit.order
        ets = orbit.eigentensors
        for i, et in enumerate(ets):
            ets[i] = et.astype(np.float64)
            ets[i] /= norm
        for of in orbit.orientation_families:
            ets = of.eigentensors
            for i, et in enumerate(ets):
                ets[i] = et.astype(np.float64)
                ets[i] /= norm

    orders = []
    for orbit in cs.orbits:
        for et in orbit.eigentensors:
            orders.append(orbit.order)

    M = cs._cvs
    for i, r in enumerate(M.row):
        M.data[i] *= cs.length_scale**orders[r]



"""Contains a calculator which given an arbitrary list of clusters and
associated force constants can calculate the energy and forces of a displaced
system
"""
import numpy as np
import math
from collections import defaultdict as dd

from hiphive.force_constants import SortedForceConstants
from ase.calculators.calculator import Calculator, all_changes
from ase.geometry import find_mic
from .numba_calc import _get_forces


class ForceConstantCalculator(Calculator):
    """This class provides an ASE calculator that can be used in conjunction
    with integrators and optimizers with the `atomic simulation environment
    (ASE) <https://wiki.fysik.dtu.dk/ase/index.html>`_. To initialize an object
    of this class one must provide the ideal atomic configuration along with a
    compatible force constant model.

    Parameters
    -----------
    fcs: ForceConstants
        the force constants instance must include the atomic configuration
    """

    implemented_properties = ['energy', 'forces']

    def __init__(self, fcs):
        Calculator.__init__(self)

        if not isinstance(fcs, SortedForceConstants):
            raise TypeError('The FC calculator requires sorted FCs.')
        self.atoms_ideal = fcs.supercell.copy()

        # Nearest neighbor distance used as maximum displacement allowed,
        # stops exploding MD simulations.
        self.max_allowed_disp = 2 * np.min(sorted(np.unique(
            self.atoms_ideal.get_all_distances(mic=True).round(4)))[1])

        self.clusters = dd(list)
        self.force_constants = dd(list)
        self.atom_indices = dd(list)
        self.atom_positions = dd(list)
        self.atom_counts = dd(list)
        self.prefactors = dd(list)
        # The main idea is to precompute the prefactor and multiplicities of
        # belonging to each cluster
        for cluster, fc in fcs.get_fc_dict().items():
            argsort = np.argsort(cluster)  # TODO: is already True?
            cluster = np.array(sorted(cluster))
            nbody = len(set(cluster))
            order = len(cluster)
            key = (nbody, order)
            self.clusters[key].append(cluster)
            assert fc.shape == (3,) * order
            self.force_constants[key].append(fc.transpose(argsort))
            unique = np.unique(cluster, return_index=True, return_counts=True)
            self.atom_indices[key].append(unique[0])
            self.atom_positions[key].append(unique[1])
            self.atom_counts[key].append(unique[2])
            prefac = [-count/np.prod(list(map(math.factorial, unique[2])))
                      for count in unique[2]]
            self.prefactors[key].append(prefac)
        for d in [self.clusters,
                  self.force_constants,
                  self.atom_indices,
                  self.atom_positions,
                  self.atom_counts,
                  self.prefactors,
                  ]:
            for k, v in d.items():
                d[k] = np.array(v)

    def calculate(self, atoms=None, properties=['energy'],
                  system_changes=all_changes):
        Calculator.calculate(self, atoms, properties, system_changes)
        self._check_atoms()
        self._compute_displacements()

        if 'forces' in properties or 'energy' in properties:
            E, forces = self.compute_energy_and_forces()
            self.results['forces'] = forces
            self.results['energy'] = E

    def _check_atoms(self):
        """Check that the atomic configuration, with which the calculator is
        associated, is compatible with the ideal configuration provided during
        initialization."""
        if len(self.atoms) != len(self.atoms_ideal):
            raise ValueError('Length of atoms does not match reference atoms')
        if not all(self.atoms.numbers == self.atoms_ideal.numbers):
            raise ValueError('Atomic numbers does not match reference atoms')

    def _compute_displacements(self):
        """Evaluate the atomic displacements between the current and the ideal
        (reference) configuration."""
        displacements = []
        for pos, ideal_pos in zip(self.atoms.positions,
                                  self.atoms_ideal.positions):
            v_ij = np.array([pos - ideal_pos])
            displacements.append(find_mic(v_ij, self.atoms.cell,
                                          pbc=True)[0][0])
        self.displacements = np.array(displacements)

        # sanity check that displacements are not too large
        max_disp = np.max(np.linalg.norm(self.displacements, axis=1))
        if max_disp > self.max_allowed_disp:
            raise ValueError(
                'Displacement {:.5f} larger than maximum allowed displacement'
                ' {:.5f}'.format(max_disp, self.max_allowed_disp))

    def compute_energy_and_forces(self):
        """Compute energy and forces.

        Returns
        -------
        float, list(list(float))
            energy and forces
        """
        E = np.zeros(1)
        forces = np.zeros((len(self.atoms), 3))

        for key in self.clusters.keys():
            nbody = key[0]
            order = key[1]
            _get_forces(self.clusters[key],
                        self.force_constants[key],
                        self.atom_indices[key],
                        self.atom_positions[key],
                        self.atom_counts[key],
                        self.prefactors[key],
                        nbody, order,
                        forces, E, self.displacements)
        return float(E), forces

    def __repr__(self):
        fc_dict_str = '{{{}: {}, ...}}'.format(
            self.clusters[0], self.force_constants[0])
        fcs_str = 'ForceConstants(fc_dict={}, atoms={!r})'.format(
            fc_dict_str, self.atoms_ideal)
        return 'ForceConstantCalculator({})'.format(fcs_str)



"""
Optimizer with cross validation score
"""

import numpy as np
from sklearn.model_selection import KFold, ShuffleSplit
from typing import Any, Dict, Tuple
from .base_optimizer import BaseOptimizer
from .optimizer import Optimizer
from .fit_methods import fit
from .tools import ScatterData


validation_methods = {
    'k-fold': KFold,
    'shuffle-split': ShuffleSplit,
}


class CrossValidationEstimator(BaseOptimizer):
    """
    This class provides an optimizer with cross validation for solving the
    linear :math:`\\boldsymbol{A}\\boldsymbol{x} = \\boldsymbol{y}` problem.
    Cross-validation (CV) scores are calculated by splitting the
    available reference data in multiple different ways.  It also produces
    the finalized model (using the full input data) for which the CV score
    is an estimation of its performance.

    Warning
    -------
    Repeatedly setting up a CrossValidationEstimator and training
    *without* changing the seed for the random number generator will yield
    identical or correlated results, to avoid this please specify a different
    seed when setting up multiple CrossValidationEstimator instances.

    Parameters
    ----------
    fit_data : tupe(numpy.ndarray, numpy.ndarray)
        the first element of the tuple represents the fit matrix `A`
        (`N, M` array) while the second element represents the vector
        of target values `y` (`N` array); here `N` (=rows of `A`,
        elements of `y`) equals the number of target values and `M`
        (=columns of `A`) equals the number of parameters
    fit_method : str
        method to be used for training; possible choice are
        "least-squares", "lasso", "elasticnet", "bayesian-ridge", "ardr",
        "rfe", "split-bregman"
    standardize : bool
        if True the fit matrix is standardized before fitting
    validation_method : str
        method to use for cross-validation; possible choices are
        "shuffle-split", "k-fold"
    n_splits : int
        number of times the fit data set will be split for the cross-validation
    check_condition : bool
        if True the condition number will be checked
        (this can be sligthly more time consuming for larger
        matrices)
    seed : int
        seed for pseudo random number generator

    Attributes
    ----------
    train_scatter_data : ScatterData
        contains target and predicted values from each individual
        traininig set in the cross-validation split;
        :class:`ScatterData` is a namedtuple.
    validation_scatter_data : ScatterData
        contains target and predicted values from each individual
        validation set in the cross-validation split;
        :class:`ScatterData` is a namedtuple.

    """

    def __init__(self,
                 fit_data: Tuple[np.ndarray, np.ndarray],
                 fit_method: str = 'least-squares',
                 standardize: bool = True,
                 validation_method: str = 'k-fold',
                 n_splits: int = 10,
                 check_condition: bool = True,
                 seed: int = 42,
                 **kwargs) -> None:

        super().__init__(fit_data, fit_method, standardize, check_condition,
                         seed)

        if validation_method not in validation_methods.keys():
            msg = ['Validation method not available']
            msg += ['Please choose one of the following:']
            for key in validation_methods:
                msg += [' * ' + key]
            raise ValueError('\n'.join(msg))
        self._validation_method = validation_method
        self._n_splits = n_splits

        self._set_kwargs(kwargs)

        # data set splitting object
        self._splitter = validation_methods[validation_method](
            n_splits=self.n_splits, random_state=seed,
            **self._split_kwargs)

        self.train_scatter_data = None
        self.validation_scatter_data = None

        self._parameters_splits = None
        self._rmse_train_splits = None
        self._rmse_valid_splits = None
        self._rmse_train_final = None

    def train(self) -> None:
        """ Constructs the final model using all input data available. """
        self._fit_results = fit(self._A, self._y, self.fit_method,
                                self.standardize, self._check_condition,
                                **self._fit_kwargs)
        self._rmse_train_final = self.compute_rmse(self._A, self._y)

    def validate(self) -> None:
        """ Runs validation. """
        train_target, train_predicted = [], []
        valid_target, valid_predicted = [], []
        rmse_train_splits, rmse_valid_splits = [], []
        parameters_splits = []
        for train_set, test_set in self._splitter.split(self._A):
            opt = Optimizer((self._A, self._y), self.fit_method,
                            standardize=self.standardize,
                            train_set=train_set,
                            test_set=test_set,
                            check_condition=self._check_condition,
                            **self._fit_kwargs)
            opt.train()

            parameters_splits.append(opt.parameters)
            rmse_train_splits.append(opt.rmse_train)
            rmse_valid_splits.append(opt.rmse_test)
            train_target.extend(opt.train_scatter_data.target)
            train_predicted.extend(opt.train_scatter_data.predicted)
            valid_target.extend(opt.test_scatter_data.target)
            valid_predicted.extend(opt.test_scatter_data.predicted)

        self._parameters_splits = np.array(parameters_splits)
        self._rmse_train_splits = np.array(rmse_train_splits)
        self._rmse_valid_splits = np.array(rmse_valid_splits)
        self.train_scatter_data = ScatterData(
            target=np.array(train_target), predicted=np.array(train_predicted))
        self.validation_scatter_data = ScatterData(
            target=np.array(valid_target), predicted=np.array(valid_predicted))

    def _set_kwargs(self, kwargs: dict) -> None:
        """
        Sets up fit_kwargs and split_kwargs.
        Different split methods need different keywords.
        """
        self._fit_kwargs = {}
        self._split_kwargs = {}

        if self.validation_method == 'k-fold':
            self._split_kwargs['shuffle'] = True  # default True
            for key, val in kwargs.items():
                if key in ['shuffle']:
                    self._split_kwargs[key] = val
                else:
                    self._fit_kwargs[key] = val
        elif self.validation_method == 'shuffle-split':
            for key, val in kwargs.items():
                if key in ['test_size', 'train_size']:
                    self._split_kwargs[key] = val
                else:
                    self._fit_kwargs[key] = val

    @property
    def summary(self) -> Dict[str, Any]:
        """ comprehensive information about the optimizer """
        info = super().summary

        # Add class specific data
        info['validation_method'] = self.validation_method
        info['n_splits'] = self.n_splits
        info['rmse_train_final'] = self.rmse_train_final
        info['rmse_train'] = self.rmse_train
        info['rmse_train_splits'] = self.rmse_train_splits
        info['rmse_validation'] = self.rmse_validation
        info['rmse_validation_splits'] = self.rmse_validation_splits
        info['train_scatter_data'] = self.train_scatter_data
        info['validation_scatter_data'] = self.validation_scatter_data

        # add kwargs used for fitting and splitting
        info = {**info, **self._fit_kwargs, **self._split_kwargs}
        return info

    def __repr__(self) -> str:
        kwargs = dict()
        kwargs['fit_method'] = self.fit_method
        kwargs['validation_method'] = self.validation_method
        kwargs['n_splits'] = self.n_splits
        kwargs['seed'] = self.seed
        kwargs = {**kwargs, **self._fit_kwargs, **self._split_kwargs}
        return 'CrossValidationEstimator((A, y), {})'.format(
            ', '.join('{}={}'.format(*kwarg) for kwarg in kwargs.items()))

    @property
    def validation_method(self) -> str:
        """ validation method name """
        return self._validation_method

    @property
    def n_splits(self) -> int:
        """ number of splits (folds) used for cross-validation """
        return self._n_splits

    @property
    def parameters_splits(self) -> np.ndarray:
        """ all parameters obtained during cross-validation """
        return self._parameters_splits

    @property
    def n_nonzero_parameters_splits(self) -> list:
        """ number of non-zero parameters for each split """
        if self.parameters_splits is None:
            return None
        else:
            return [np.count_nonzero(p) for p in self.parameters_splits]

    @property
    def rmse_train_final(self) -> float:
        """
        root mean squared error when using the full set of input data
        """
        return self._rmse_train_final

    @property
    def rmse_train(self) -> float:
        """
        average root mean squared training error obtained during
        cross-validation
        """
        if self._rmse_train_splits is None:
            return None
        return np.sqrt(np.mean(self._rmse_train_splits**2))

    @property
    def rmse_train_splits(self) -> np.ndarray:
        """
        root mean squared training errors obtained during
        cross-validation
        """
        return self._rmse_train_splits

    @property
    def rmse_validation(self) -> float:
        """ average root mean squared cross-validation error """
        if self._rmse_valid_splits is None:
            return None
        return np.sqrt(np.mean(self._rmse_valid_splits**2))

    @property
    def rmse_validation_splits(self) -> np.ndarray:
        """
        root mean squared validation errors obtained during
        cross-validation
        """
        return self._rmse_valid_splits



"""
Functionality for enforcing rotational sum rules
"""
from sklearn.linear_model import Ridge
import numpy as np
from scipy.sparse import coo_matrix
from .utilities import SparseMatrix


def enforce_rotational_sum_rules(cs, parameters, sum_rules, **kwargs):
    """ Enforces rotational sum rules by projecting parameters.

    Note
    ----
    The interface to this function might change in future releases.

    Parameters
    ----------
    cs : ClusterSpace
        the underlying cluster space
    parameters : numpy.ndarray
        parameters to be constrained
    sum_rules : list(str)
        type of sum rules to enforce; possible values: 'Huang', 'Born-Huang'
    ridge_alpha : float
        hyperparameter to the ridge regression algorithm; keyword argument
        passed to the optimizer; larger values specify stronger regularization,
        i.e. less correction but higher stability [default: 1e-6]
    iterations : int
        number of iterations to run the projection since each step projects
        the solution down to each nullspace in serial; keyword argument
        passed to the optimizer [default: 10]

    Returns
    -------
    numpy.ndarray
        constrained parameters

    Examples
    --------
    The rotational sum rules can be enforced to the parameters before
    constructing a force constant potential as illustrated by the following
    snippet::

        cs = ClusterSpace(reference_structure, cutoffs)
        sc = StructureContainer(cs)
        # add structures to structure container
        opt = Optimizer(sc.get_fit_data())
        opt.train()
        new_params = enforce_rotational_sum_rules(cs, opt.parameters,
            sum_rules=['Huang', 'Born-Huang'])
        fcp = ForceConstantPotential(cs, new_params)
    """
    parameters = parameters.copy()
    # handle optional kwargs input
    iterations = kwargs.pop('iterations', 10)
    ridge_alpha = kwargs.pop('ridge_alpha', 1e-6)
    if len(kwargs) != 0:
        raise ValueError('Unrecognised kwargs: {}'.format(kwargs))

    # make orbit-parameter index map
    params = []
    n = 0
    for orbit_index, orbit in enumerate(cs.orbits):
        number_params_in_orbit = len(orbit.eigentensors)
        params.append(list(range(n, n + number_params_in_orbit)))
        n += number_params_in_orbit

    # create lookuptable for force constants
    lookup = {}
    for orbit_index, orbit in enumerate(cs.orbits):
        for of in orbit.orientation_families:
            for cluster_index, perm_index in zip(of.cluster_indices, of.permutation_indices):
                cluster = cs._cluster_list[cluster_index]
                perm = cs._permutations[perm_index]
                lookup[tuple(cluster)] = [et.transpose(perm) for et in of.eigentensors], orbit_index

    # append the sum rule matrices
    Ms = []
    args = (lookup, params, cs.atom_list, cs._prim)
    for sum_rule in sum_rules:
        if sum_rule == 'Huang':
            Ms.append(_create_Huang_constraint(*args))
        elif sum_rule == 'Born-Huang':
            Ms.append(_create_Born_Huang_constraint(*args))
        else:
            raise ValueError('At least one unknown sum rule: {}'.format(sum_rule))

    cvs_trans = cs._cvs

    for i, M in enumerate(Ms):
        row, col, data = [], [], []
        for r, c, v in M.row_list():
            row.append(r)
            col.append(c)
            data.append(np.float64(v))
        M = coo_matrix((data, (row, col)), shape=M.shape)
        Ms[i] = M.dot(cvs_trans)

    for _ in range(iterations):
        for M in Ms:
            clf = Ridge(alpha=ridge_alpha)
            d = M.dot(parameters)
            clf.fit(M, d)
            parameters -= clf.coef_  # subtract the correction

    return parameters


def _create_Huang_constraint(lookup, parameter_map, atom_list, prim):

    m = SparseMatrix(3**4, parameter_map[-1][-1] + 1, 0)

    def R(i, j):
        pi = atom_list[i].pos(prim.basis, prim.cell)
        pj = atom_list[j].pos(prim.basis, prim.cell)
        return pi - pj

    for i in range(len(prim)):
        for j in range(len(atom_list)):
            ets, orbit_index = lookup.get(tuple(sorted((i, j))), (None, None))
            if ets is None:
                continue
            inv_perm = np.argsort(np.argsort((i, j)))
            et_indices = parameter_map[orbit_index]
            for et, et_index in zip(ets, et_indices):
                et = et.transpose(inv_perm)
                Rij = R(i, j)
                Cij = np.einsum(et, [0, 1], Rij, [2], Rij, [3])
                Cij -= Cij.transpose([2, 3, 0, 1])
                for k in range(3**4):
                    m[k, et_index] += Cij.flat[k]
    return m


def _create_Born_Huang_constraint(lookup, parameter_map, atom_list, prim):

    constraints = []

    for i in range(len(prim)):
        m = SparseMatrix(3**3, parameter_map[-1][-1] + 1, 0)
        for j in range(len(atom_list)):
            ets, orbit_index = lookup.get(tuple(sorted((i, j))), (None, None))
            if ets is None:
                continue
            inv_perm = np.argsort(np.argsort((i, j)))
            et_indices = parameter_map[orbit_index]
            R = atom_list[j].pos(prim.basis, prim.cell)
            for et, et_index in zip(ets, et_indices):
                et = et.transpose(inv_perm)
                tmp = np.einsum(et, [0, 1], R, [2])
                tmp -= tmp.transpose([0, 2, 1])
                for k in range(3**3):
                    m[k, et_index] += tmp.flat[k]
        constraints.append(m)

    M = SparseMatrix.vstack(*constraints)
    return M



import numpy as np
from ase.units import kB
from hiphive.io.logging import logger

logger = logger.getChild(__name__)


def generate_phonon_rattled_structures(atoms, fc2, n_structures, temperature):
    """
    Returns list of phonon-rattled configurations.

    Configurations are generated by superimposing harmonic phonon
    eigenmodes with random amplitudes and phase factors consistent
    with a certain temperature.

    Let :math:`\\boldsymbol{X}_{ai}` be the phonon modes indexed by atom
    :math:`a` and mode :math:`i`, :math:`\\omega_i` the phonon
    frequencies, and let :math:`0 < Q_i \\leq 1` and :math:`0 \\leq
    U_i < 1` be uniformly random numbers.  Then

    .. math::

        \\boldsymbol{R}_a
        = \\boldsymbol{R}^0_a
        + \\left<\\frac{k_B T}{m_a} \\right>^{1/2}
        \\sum_i \\frac{1}{\\omega_i} \\boldsymbol{X}_{ai}
        \\sqrt{-2 \\ln Q_i} \\cos(\\pi \\omega_i U_i)

    See: West and Estreicher PRL 96, 115504 (2006)

    Parameters
    ----------
    atoms : ase.Atoms
        prototype structure
    fc2 : numpy.ndarray
        second order force constant matrix, with shape `(3N, 3N)` or
        `(N, N, 3, 3)`. The conversion will be done internally if.
    n_structures : int
        number of structures to generate
    temperature : float
        temperature in Kelvin

    Returns
    -------
    list(ase.Atoms)
        generated structures
    """
    structures = []
    pr = _PhononRattler(atoms.get_masses(), fc2)
    for _ in range(n_structures):
        atoms_tmp = atoms.copy()
        pr(atoms_tmp, temperature)
        structures.append(atoms_tmp)
    return structures


def _phonon_rattle(m_a, T, w2_s, e_sai):
    """ Thermal excitation of phonon modes as described by West and
    Estreicher PRL 96, 115504 (2006).

    _s is a mode index
    _i is a Carteesian index
    _a is an atom index

    Parameters
    ----------
    m_a : numpy.ndarray
        masses (N,)
    T : float
        temperature in Kelvin
    w2_s : numpy.ndarray
        the squared frequencies from the igenvalue problem (3*N,)
    e_sai : numpy.ndarray
        polarizations (3*N, N, 3)

    Returns
    -------
    displacements : numpy.ndarray
        shape (N, 3)
    """
    n_modes = 3 * len(m_a)
    prefactor_a = np.sqrt(2 * kB * T / m_a)

    # The three modes closest to zero are assumed to be zero, i.e. acoustic sum rules are assumed
    argsort = np.argsort(np.abs(w2_s))
    w2_gamma = w2_s[argsort][:3]
    e_sai = e_sai[argsort][3:]
    w2_s = w2_s[argsort][3:]
    if np.any(np.abs(w2_gamma) > 1e-6):
        logger.warning('Acoustic sum rules not enforced, squared frequencies: {}'.format(w2_gamma))

    # treat imaginary modes as real
    if np.any(w2_s < 0):
        logger.warning('Imaginary modes present')
    w_s = np.sqrt(np.abs(w2_s))

    phases_s = np.random.uniform(0, 2 * np.pi, size=n_modes - 3)
    amplitudes_s = np.sqrt(-np.log(1 - np.random.random(n_modes - 3)))

    # prefactors are reshaped in order for numpy broadcast to work
    prefactor_a = prefactor_a.reshape(-1, 1)
    u_ai = prefactor_a * np.tensordot(amplitudes_s * np.cos(phases_s) / w_s, e_sai, (0, 0))
    return u_ai  # displacements


class _PhononRattler:
    """
    Class to be able to conveniently save modes and frequencies needed
    for phonon rattle.

    Parameters
    ----------
    masses : numpy.ndarray
        masses (N,)
    force_constants : numpy.ndarray
        second order force constant matrix, with shape `(3N, 3N)` or
        `(N, N, 3, 3)`. The conversion will be done internally if.
    """
    def __init__(self, masses, force_constants):
        n_atoms = len(masses)
        if len(force_constants.shape) == 4:  # assume shape = (n_atoms, n_atoms, 3, 3)
            force_constants = force_constants.transpose(0, 2, 1, 3)
            force_constants = force_constants.reshape(3 * n_atoms, 3 * n_atoms)
            # Now the fc should have shape = (n_atoms * 3, n_atoms * 3)
        # Construct the dynamical matrix
        inv_root_masses = (1 / np.sqrt(masses)).repeat(3)
        D = np.outer(inv_root_masses, inv_root_masses)
        D *= force_constants
        # find modes and energies
        w2_s, e_sai = np.linalg.eigh(D)
        # reshape to get atom index and Cartesian index separate
        e_sai = e_sai.T.reshape(-1, n_atoms, 3)
        self.w2_s = w2_s
        self.e_sai = e_sai
        self.masses = masses

    def __call__(self, atoms, T):
        """ rattle atoms by adding displacements

        Parameters
        ----------
        atoms : ase.Atoms
            Ideal structure to add displacements to.
        T : float
            temperature in Kelvin
        """
        u_ai = _phonon_rattle(self.masses, T, self.w2_s, self.e_sai)
        atoms.positions += u_ai



"""
This module provides functionality for storing structures and their fit
matrices together with target forces and displacements
"""

import tarfile
import numpy as np
from collections import OrderedDict
from ase.calculators.singlepoint import SinglePointCalculator

from .io.read_write_files import (add_items_to_tarfile_hdf5,
                                  add_items_to_tarfile_pickle,
                                  add_items_to_tarfile_custom,
                                  add_list_to_tarfile_custom,
                                  read_items_hdf5,
                                  read_items_pickle,
                                  read_list_custom)

from .cluster_space import ClusterSpace
from .force_constant_model import ForceConstantModel
from .io.logging import logger
logger = logger.getChild('structure_container')


class StructureContainer:
    """
    This class serves as a container for structures as well as associated
    fit properties and fit matrices.

    Parameters
    -----------
    cs : ClusterSpace
        cluster space that is the basis for the container
    fit_structure_list : list(FitStructure)
         structures to be added to the container
    """

    def __init__(self, cs, fit_structure_list=None):
        """
        Attributes
        -----------
        _cs : ClusterSpace
            cluster space that is the basis for the container
        _structure_list : list(FitStructure)
            structures to add to container
        _previous_fcm : ForceConstantModel
            FCM object used for last fit matrix calculation; check will be
            carried out to decide if this FCM can be used for a new structure
            or not, which often enables a considerable speed-up
        """
        self._cs = cs.copy()
        self._previous_fcm = None

        # Add atoms from atoms_list
        self._structure_list = []
        if fit_structure_list is not None:
            for fit_structure in fit_structure_list:
                if not isinstance(fit_structure, FitStructure):
                    raise TypeError('Can only add FitStructures')
                self._structure_list.append(fit_structure)

    def __len__(self):
        return len(self._structure_list)

    def __getitem__(self, ind):
        return self._structure_list[ind]

    @property
    def data_shape(self):
        """ tuple : tuple of integers representing the shape of the fit data
        matrix """
        n_cols = self._cs.n_dofs
        n_rows = sum(len(fs) * 3 for fs in self)
        if n_rows == 0:
            return None
        return n_rows, n_cols

    @property
    def cluster_space(self):
        """ ClusterSpace : copy of the cluster space the structure
        container is based on"""
        return self._cs.copy()

    @staticmethod
    def read(fileobj, read_structures=True):
        """Restore a StructureContainer object from file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to load from (file object)
        read_structures : bool
            if True the structures will be read; if False only the cluster
            space will be read
        """
        if isinstance(fileobj, str):
            tar_file = tarfile.open(mode='r', name=fileobj)
        else:
            tar_file = tarfile.open(mode='r', fileobj=fileobj)

        # Read clusterspace
        f = tar_file.extractfile('cluster_space')
        cs = ClusterSpace.read(f)

        # Read fitstructures
        fit_structure_list = None
        if read_structures:
            fit_structure_list = read_list_custom(tar_file, 'fit_structure', FitStructure.read)

        # setup StructureContainer
        sc = StructureContainer(cs, fit_structure_list)

        # Read previous FCM if it exists
        if 'previous_fcm' in tar_file.getnames():
            f = tar_file.extractfile('previous_fcm')
            fcm = ForceConstantModel.read(f)
            sc._previous_fcm = fcm

        return sc

    def write(self, f):
        """Write a StructureContainer instance to a file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """

        if isinstance(f, str):
            tar_file = tarfile.open(mode='w', name=f)
        else:
            tar_file = tarfile.open(mode='w', fileobj=f)

        # save cs and previous_fcm (if it exists)
        custom_items = dict(cluster_space=self._cs)
        if self._previous_fcm is not None:
            custom_items['previous_fcm'] = self._previous_fcm
        add_items_to_tarfile_custom(tar_file, custom_items)

        # save fit structures
        add_list_to_tarfile_custom(tar_file, self._structure_list, 'fit_structure')

        tar_file.close()

    def add_structure(self, atoms, **meta_data):
        """Add a structure to the container.

        Note that custom information about the atoms object may not be
        stored inside, for example an ASE
        :class:`SinglePointCalculator` will not be kept.

        Parameters
        ----------
        atoms : ase.Atoms
            the structure to be added; the Atoms object must contain
            supplementary per-atom arrays with displacements and forces
        meta_data : dict
            dict with meta_data about the atoms
        """

        atoms_copy = atoms.copy()

        # atoms object must contain displacements
        if 'displacements' not in atoms_copy.arrays.keys():
            raise ValueError('Atoms must have displacements array')

        # atoms object must contain forces
        if 'forces' not in atoms_copy.arrays.keys():
            if isinstance(atoms.calc, SinglePointCalculator):
                atoms_copy.new_array('forces', atoms.get_forces())
            else:
                raise ValueError('Atoms must have forces')

        # check if an identical atoms object already exists in the container
        for i, structure in enumerate(self._structure_list):
            if are_configurations_equal(atoms_copy, structure.atoms):
                raise ValueError('Atoms is identical to structure {}'.format(i))

        logger.info('Adding structure')
        M = self._compute_fit_matrix(atoms)
        structure = FitStructure(atoms_copy, M, **meta_data)
        self._structure_list.append(structure)

    def delete_all_structures(self):
        """ Remove all current structures in StructureContainer. """
        self._structure_list = []

    def get_fit_data(self, structures=None):
        """Return fit data for structures. The fit matrices and target forces
        for the structures are stacked into NumPy arrays.

        Parameters
        ----------
        structures: list, tuple
            list of integers corresponding to structure indices. Defaults to
            None and in that case returns all fit data available.

        Returns
        -------
        numpy.ndarray, numpy.ndarray
            stacked fit matrices, stacked target forces for the structures
        """
        if structures is None:
            structures = range(len(self))

        M_list, f_list = [], []
        for i in structures:
            M_list.append(self._structure_list[i].fit_matrix)
            f_list.append(self._structure_list[i].forces.flatten())

        if len(M_list) == 0:
            return None
        return np.vstack(M_list), np.hstack(f_list)

    def __str__(self):
        if len(self._structure_list) > 0:
            return self._get_str_structure_list()
        else:
            return 'Empty StructureContainer'

    def __repr__(self):
        return 'StructureContainer({!r}, {!r})'.format(
            self._cs, self._structure_list)

    def _get_str_structure_list(self):
        """ Return formatted string of the structure list """
        def str_structure(index, structure):
            fields = OrderedDict([
                ('index',     '{:^4}'.format(index)),
                ('num-atoms', '{:^5}'.format(len(structure))),
                ('avg-disp',  '{:7.4f}'
                 .format(np.mean([np.linalg.norm(d) for d in
                                  structure.displacements]))),
                ('avg-force', '{:7.4f}'
                 .format(np.mean([np.linalg.norm(f) for f in
                                  structure.forces]))),
                ('max-force', '{:7.4f}'
                 .format(np.max([np.linalg.norm(f) for f in
                                 structure.forces])))])
            s = []
            for name, value in fields.items():
                n = max(len(name), len(value))
                if index < 0:
                    s += ['{s:^{n}}'.format(s=name, n=n)]
                else:
                    s += ['{s:^{n}}'.format(s=value, n=n)]
            return ' | '.join(s)

        # table width
        dummy = self._structure_list[0]
        n = len(str_structure(-1, dummy))

        # table header
        s = []
        s.append(' Structure Container '.center(n, '='))
        s += ['{:22} : {}'.format('Total number of structures', len(self))]
        s.append(''.center(n, '-'))
        s.append(str_structure(-1, dummy))
        s.append(''.center(n, '-'))

        # table body
        for i, structure in enumerate(self._structure_list):
            s.append(str_structure(i, structure))
        s.append(''.center(n, '='))
        return '\n'.join(s)

    def _compute_fit_matrix(self, atoms):
        """ Compute fit matrix for a single atoms object """
        logger.info('Computing fit matrix')
        if atoms != getattr(self._previous_fcm, 'atoms', None):
            logger.info('  Building new FCM object')
            self._previous_fcm = ForceConstantModel(atoms, self._cs)
        else:
            logger.info('  Reusing old FCM object')
        return self._previous_fcm.get_fit_matrix(atoms.get_array('displacements'))


class FitStructure:
    """This class holds a structure with displacements and forces as well as
    the fit matrix.

    Parameters
    ----------
    atoms : ase.Atoms
        supercell structure
    fit_matrix : numpy.ndarray
        fit matrix, `N, M` array with `N = 3 * len(atoms)`
    meta_data : dict
        any meta data that needs to be stored in the FitStructure
    """

    def __init__(self, atoms, fit_matrix, **meta_data):
        if 3 * len(atoms) != fit_matrix.shape[0]:
            raise ValueError('fit matrix not compatible with atoms')
        self._atoms = atoms
        self._fit_matrix = fit_matrix
        self.meta_data = meta_data

    @property
    def fit_matrix(self):
        """ numpy.ndarray : the fit matrix """
        return self._fit_matrix

    @property
    def atoms(self):
        """ ase.Atoms : supercell structure """
        return self._atoms.copy()

    @property
    def forces(self):
        """ numpy.ndarray : forces """
        return self._atoms.get_array('forces')

    @property
    def displacements(self):
        """ numpy.ndarray : atomic displacements """
        return self._atoms.get_array('displacements')

    def __getattr__(self, key):
        """ Accesses meta_data if possible and returns value. """
        if key not in self.meta_data.keys():
            return super().__getattribute__(key)
        return self.meta_data[key]

    def __len__(self):
        return len(self._atoms)

    def __str__(self):
        s = []
        s.append(' FitStructure '.center(65, '='))
        s.append('Formula: {}'.format(self.atoms.get_chemical_formula()))
        s.append(('Cell:' + '\n   [{:9.5f} {:9.5f} {:9.5f}]'*3).format(
            *self.atoms.cell[0], *self.atoms.cell[1], *self.atoms.cell[2]))
        s.append('Atoms (positions, displacements, forces):')
        for atom, disp, force in zip(self.atoms, self.displacements, self.forces):
            array_fmt = '[ {:9.5f} {:9.5f} {:9.5f} ]'
            row_str = '{:3d} {}'.format(atom.index, atom.symbol)
            row_str += array_fmt.format(*atom.position)
            row_str += array_fmt.format(*disp)
            row_str += array_fmt.format(*force)
            s.append(row_str)
        return '\n'.join(s)

    def __repr__(self):
        return 'FitStructure({!r}, ..., {!r})'.format(self.atoms, self.meta_data)

    def write(self, fileobj):
        """ Write the instance to file.

        Parameters
        ----------
        fileobj : str or file object
            name of input file (str) or stream to write to (file object)
        """
        if isinstance(fileobj, str):
            tar_file = tarfile.open(name=fileobj, mode='w')
        else:
            tar_file = tarfile.open(fileobj=fileobj, mode='w')

        items_pickle = dict(atoms=self._atoms, meta_data=self.meta_data)
        items_hdf5 = dict(fit_matrix=self.fit_matrix)

        add_items_to_tarfile_pickle(tar_file, items_pickle, 'items.pickle')
        add_items_to_tarfile_hdf5(tar_file, items_hdf5, 'fit_matrix.hdf5')

        tar_file.close()

    @staticmethod
    def read(fileobj, read_fit_matrix=True):
        """ Read a OrientationFamily instance from a file.

        Parameters
        ----------
        fileobj : str or file object
            name of input file (str) or stream to read from (file object)
        read_fit_matrix : bool
            whether or not to read the fit_matrix
        Returns
        -------
        FitStructure instance
        """
        if isinstance(fileobj, str):
            tar_file = tarfile.open(mode='r', name=fileobj)
        else:
            tar_file = tarfile.open(mode='r', fileobj=fileobj)

        items = read_items_pickle(tar_file, 'items.pickle')
        fit_matrix = read_items_hdf5(tar_file, 'fit_matrix.hdf5')['fit_matrix']

        return FitStructure(items['atoms'], fit_matrix, **items['meta_data'])


def are_configurations_equal(atoms1, atoms2, tol=1e-10):
    """ Compare if two configurations are equal within some tolerance. This
    includes checking all available arrays in the two atoms objects.

    Parameters
    ----------
    atoms1 : ase.Atoms
    atoms2 : ase.Atoms

    Returns
    -------
    bool
        True if atoms are equal, False otherwise
    """

    # pbc
    if not all(atoms1.pbc == atoms2.pbc):
        return False

    # cell
    if not np.allclose(atoms1.cell, atoms2.cell, atol=tol, rtol=0.0):
        return False

    # arrays
    if not len(atoms1.arrays.keys()) == len(atoms2.arrays.keys()):
        return False
    for key, array1 in atoms1.arrays.items():
        if key not in atoms2.arrays.keys():
            return False
        if not np.allclose(array1, atoms2.arrays[key], atol=tol, rtol=0.0):
            return False

    # passed all test, atoms must be equal
    return True



"""
Contains the force constant modell (FCM) which handles cluster and force
constant information of a super cell object.
"""
import math
import pickle
import numpy as np

from .cluster_space import ClusterSpace
from .core.atoms import Atom, atom_to_spos, spos_to_atom
from .core.structure_alignment import align_supercell
from .core.orbits import Orbit, OrientationFamily
from .core.tensors import (rotation_to_cart_coord,
                           rotate_tensor_precalc, rotation_tensor_as_matrix)
from .core.utilities import Progress, BiMap
from .force_constants import SortedForceConstants
from .io.logging import logger
from .calculators.numba_calc import (clusters_force_contribution,
                                     cluster_force_contribution)
from scipy.sparse import coo_matrix, vstack

logger = logger.getChild('fcm')


class ForceConstantModel:
    """Transfers a cluster space onto a super structure

    Contains the full description of all clusters and the force constants
    within a super cell with periodic boundary conditions.

    Parameters
    ----------
    atoms : ase.Atoms
        configuration to which the cluster space is to be applied
    cs : ClusterSpace
        a cluster space compatible with the structure of the atoms
    """
    def __init__(self, atoms, cs):

        self.atoms = atoms.copy()
        self.orbits = []
        self.cluster_list = BiMap()
        self.cs = cs
        self._populate_orbits(atoms)

    # TODO: refactor
    def _populate_orbits(self, atoms):
        """Map the orbits from the underlying force constant potential onto the
        supercell structure associated with this force constant model.

        """
        # TODO: Comment function
        atom_lookup = {}
        cs = self.cs

        aligned_super_cell, scR, _ = align_supercell(
            atoms, cs.primitive_structure, cs.symprec)
        sc = aligned_super_cell.copy()
        sc.pbc = False
        sc.cell = cs.primitive_structure.cell
        atom_list = BiMap()
        # TODO: fix this. see also 334d8572
        for spos in sc.get_scaled_positions():
            atom_list.append(spos_to_atom(spos, cs.primitive_structure.basis,
                                          cs.symprec))
        sorted_cluster_list = BiMap()
        if hasattr(cs, 'rotation_matrices'):
            rotations = []
            for R in cs.rotation_matrices:
                rotations.append(rotation_to_cart_coord(
                    R, cs.primitive_structure.cell))

        def get_atom_index(atom):
            tupd_atom = (atom.site, *atom.offset)
            if tupd_atom in atom_lookup:
                return atom_lookup[tupd_atom]
            spos = atom_to_spos(atom, cs.primitive_structure.basis)
            pos = np.dot(spos, cs.primitive_structure.cell)
            spos = np.dot(pos, np.linalg.inv(aligned_super_cell.cell))
            atom = spos_to_atom(spos, aligned_super_cell.basis,
                                self.cs.symprec)
            atom_lookup[tupd_atom] = atom.site
            return atom.site

        def get_mapped_cluster(cluster, offset):
            new_cluster = []
            for atom_index in cluster:
                atom = cs.atom_list[atom_index]
                translated_atom = Atom(atom.site, np.add(atom.offset, offset))
                index = get_atom_index(translated_atom)
                new_cluster.append(index)
            return new_cluster
        logger.info('Populating orbits')
        bar = Progress(len(cs.orbits))
        scR_tensormatrix_lookup = dict()
        R_inv_tensormatrix_lookup = dict()
        for orbit_index, orbit in enumerate(cs.orbits):

            new_orbit = Orbit()
            new_orbit.order = orbit.order

            scR_tensormatrix = scR_tensormatrix_lookup.get(orbit.order, None)
            if scR_tensormatrix is None:
                scR_tensormatrix = rotation_tensor_as_matrix(scR, orbit.order)
                scR_tensormatrix_lookup[orbit.order] = scR_tensormatrix

            if len(orbit.eigentensors) > 0:
                ets = []
                for et in orbit.eigentensors:
                    ets.append(rotate_tensor_precalc(et, scR_tensormatrix))
                new_orbit.eigentensors = ets
                new_orbit.force_constant = np.zeros(ets[0].shape)
            else:
                new_orbit.force_constant = rotate_tensor_precalc(
                        orbit.force_constant, scR_tensormatrix)
            cluster = cs.cluster_list[orbit.prototype_index]
            _, pos, counts = np.unique(np.array(cluster),
                                       return_index=True, return_counts=True)
            new_orbit.positions = pos
            prefactor = -1 / np.prod(list(map(math.factorial, counts)))
            new_orbit.prefactors = np.array([prefactor * c for c in counts])

            for of in orbit.orientation_families:

                new_of = OrientationFamily()
                if len(orbit.eigentensors) > 0:
                    R_inv_tensormatrix_index = (of.symmetry_index, orbit.order)
                    R_inv_tensormatrix = \
                        R_inv_tensormatrix_lookup.get(R_inv_tensormatrix_index,
                                                      None)
                    if R_inv_tensormatrix is None:
                        R_inv = rotations[of.symmetry_index].T
                        R_inv_tensormatrix = rotation_tensor_as_matrix(
                                R_inv, orbit.order)
                        R_inv_tensormatrix_lookup[R_inv_tensormatrix_index] = \
                            R_inv_tensormatrix
                    ets = []
                    for et in orbit.eigentensors:
                        et_of = rotate_tensor_precalc(et, R_inv_tensormatrix)
                        ets.append(rotate_tensor_precalc(
                            et_of, scR_tensormatrix))
                    new_of.eigentensors = ets
                    new_of.force_constant = np.zeros(ets[0].shape)
                else:
                    new_of.force_constant = rotate_tensor_precalc(
                            of.force_constant, scR_tensormatrix)

                cluster = cs.cluster_list[of.cluster_indices[0]]
                if isinstance(cs, ClusterSpace):
                    perm = cs._permutations[of.permutation_indices[0]]
                    cluster = [cluster[i] for i in np.argsort(perm)]
                for atom in atom_list:
                    if not atom.site == cs.atom_list[cluster[0]].site:
                        continue
                    offset = atom.offset
                    new_cluster = tuple(get_mapped_cluster(cluster, offset))
                    sorted_new_cluster = tuple(sorted(new_cluster))
                    if sorted_new_cluster in sorted_cluster_list:
                        raise Exception('Found cluster {} twice, check '
                                        'cutoff!'.format(sorted_new_cluster))
                    sorted_cluster_list.append(sorted_new_cluster)
                    self.cluster_list.append(new_cluster)
                    new_cluster_index = len(self.cluster_list) - 1
                    new_of.cluster_indices.append(new_cluster_index)
                new_orbit.orientation_families.append(new_of)
            self.orbits.append(new_orbit)
            bar.tick()
        bar.close()
        if isinstance(cs, ClusterSpace):
            self._parameters = np.zeros(self.cs.n_dofs)

    def get_force_constants(self):
        """Returns the force constants of the super cell.

        Returns
        -------
        force constants : ForceConstants
            complete set of force constants
        """

        fc_dict = dict()
        for orbit in self.orbits:
            for of in orbit.orientation_families:
                fc = of.force_constant.copy()
                for cluster_index in of.cluster_indices:
                    cluster = self.cluster_list[cluster_index]
                    perm = np.argsort(cluster)
                    sorted_cluster = tuple(sorted(cluster))
                    fc_dict[sorted_cluster] = fc.transpose(perm)
        return SortedForceConstants(fc_dict, self.atoms)

    @property
    def parameters(self):
        """ list(float): parameters """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        self._parameters = parameters
        mapped_parameters = self.cs._map_parameters(parameters)
        p = 0
        for orb in self.orbits:
            fc_is_zero = np.allclose(orb.force_constant, 0)
            params_is_zero = np.allclose(
                mapped_parameters[p: p+len(orb.eigentensors)], 0)
            if fc_is_zero and params_is_zero:
                p += len(orb.eigentensors)
                continue
            orb.force_constant *= 0
            if not params_is_zero:
                for et, a in zip(orb.eigentensors, mapped_parameters[p:]):
                    orb.force_constant += et * a
            for of in orb.orientation_families:
                of.force_constant *= 0
                if not params_is_zero:
                    for et, a in zip(of.eigentensors, mapped_parameters[p:]):
                        of.force_constant += et * a
            p += len(orb.eigentensors)

    def get_forces(self, displacements):
        """ Returns the forces in the system given displacements.

        The parameters of the model must be set to get any result.

        Parameters
        ----------
        displacements : numpy.ndarray
            displacements of each atom in the supercell (`N, 3` array)
        """
        F = np.zeros(displacements.shape)
        f = np.zeros(3)
        for orbit in self.orbits:
            if np.allclose(orbit.force_constant, 0):
                continue
            order = orbit.order
            positions = orbit.positions
            prefactors = orbit.prefactors
            for of in orbit.orientation_families:
                fc = of.force_constant.flatten()
                fc_tmp = fc.copy()

                for cluster_index in of.cluster_indices:
                    cluster = self.cluster_list[cluster_index]
                    cluster_force_contribution(
                            positions, prefactors, len(prefactors),
                            fc_tmp, fc, order,
                            displacements,
                            cluster, f, F)
        return F

    def get_fit_matrix(self, displacements):
        """ Returns the matrix used to fit the parameters.

        Represents the linear relation between the parameters and the forces.

        Parameters
        ----------
        displacements : numpy.ndarrya
            displacements of each atom in the supercell (`(N, 3)` array)
        """

        F = np.zeros(displacements.shape)
        f = np.zeros(3)
        M = np.zeros((F.size, self.cs._cvs.shape[0]))
        et_index = 0
        bar = Progress(len(self.orbits))
        for orbit in self.orbits:
            bar.tick()
            fc_tmp = np.zeros(3**orbit.order)
            for of in orbit.orientation_families:
                clusters = np.zeros((len(of.cluster_indices), orbit.order),
                                    dtype=np.int64)
                for i, cluster_index in enumerate(of.cluster_indices):
                    clusters[i] = self.cluster_list[cluster_index]
                for i, et in enumerate(of.eigentensors):
                    F[:] = 0
                    clusters_force_contribution(orbit.positions,
                                                orbit.prefactors,
                                                len(orbit.prefactors),
                                                fc_tmp,
                                                et.ravel(),
                                                orbit.order,
                                                displacements,
                                                clusters, f, F)
                    M[:, et_index + i] += F.flat
            et_index += len(orbit.eigentensors)
        M = M.dot(self.cs._cvs.toarray())
        bar.close()
        return M

    def get_fcs_sensing(self, fcs):
        """ Creates a fit matrix from force constants directly.

        If the underlying cluster space can completely span the force constants
        the correct parameters should be recovered. The method assumes that the
        force constants obey crystal, lattice and label symmetries and will
        naively incorporate only one force constant per orbit into the sensing
        matrix.

        The parameters can be extracted using e.g. least squares from numpy

        >>> parameters = np.linalg.lstsq(*fcm.get_fcs_sensing(fcs))[0]

        Parameters
        ----------
        fcs: ForceConstants
            Must be compatible with the ideal structure of the model

        Returns
        -------
        M : np.array
            The sensing matrix
        F : np.array
            The flattened, irreducible force constants
        """
        M, F = [], []
        et_index = 0
        n_parameters = self.cs._cvs.shape[0]
        for oi, orbit in enumerate(self.orbits):
            cluster = self.cluster_list[orbit.prototype_index]
            fc = fcs[cluster].ravel()
            row, col, data = [], [], []
            for i, et in enumerate(orbit.eigentensors):
                for r, d in enumerate(et.flat):
                    row.append(r)
                    col.append(et_index + i)
                    data.append(d)
            F.append(fc)
            m = coo_matrix((data, (row, col)), shape=(3**orbit.order, n_parameters))
            M.append(m)
            et_index += len(orbit.eigentensors)
        M = vstack(M)
        M = M.dot(self.cs._cvs).toarray()
        F = np.concatenate(F)
        return M, F

    @staticmethod
    def read(f):
        """Reads a force constant model from file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to load from (file object)

        Returns
        -------
        ForceConstantModel
            the original object as stored in the file
        """
        if isinstance(f, str):
            with open(f, 'rb') as fobj:
                return pickle.load(fobj)
        else:
            return pickle.load(f)

    def write(self, f):
        """Writes a force constant model to file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """
        if isinstance(f, str):
            with open(f, 'wb') as fobj:
                pickle.dump(self, fobj)
        else:
            pickle.dump(self, f)



"""
Contains the ClusterSpace object central to hiPhive
"""

import numpy as np
import tarfile

from ase.data import chemical_symbols
from collections import OrderedDict
from copy import deepcopy

from .core.cluster_space_builder import build_cluster_space
from .core.atoms import Atoms
from .core.orbits import Orbit
from .io.logging import logger
from .io.pretty_table_prints import print_table
from .io.read_write_files import (add_items_to_tarfile_pickle,
                                  add_items_to_tarfile_custom,
                                  add_list_to_tarfile_custom,
                                  read_items_pickle,
                                  read_list_custom)
from .cutoffs import Cutoffs, CutoffMaximumBody, BaseClusterFilter

from .config import Config
logger = logger.getChild('ClusterSpace')


class ClusterSpace:
    # TODO: Inherit from base class to ease the use of FCP
    """Primitive object for handling clusters and force constants of a structure.

    Parameters
    ----------
    prototype_structure : ase.Atoms
        prototype structure; spglib will be used to find a suitable cell based
        on this structure unless the cell is already a primitive cell.
    cutoffs : list or Cutoffs
        cutoff radii for different orders starting with second order
    cluster_filter : ClusterFilter
        accepts a subclass of hiphive.filters.BaseClusterFilter to further
        control which orbits to include.
    config : Config object
        config object containing information on how the cluster space should
        be built, e.g. values for tolerances and if acoustic sum rules be
        enforced

    If not config is given then kwargs can be used to specify:
    acoustic_sum_rules : bool
        If True the aucostic sum rules will be enforced by constraining the
        parameters.
    symprec : float
        numerical precision that will be used for analyzing the symmetry (this
        parameter will be forwarded to
        `spglib <https://atztogo.github.io/spglib/>`_)
    length_scale : float
        This will be used as a normalization constant for the eigentensors

    Examples
    --------

    To instantiate a :class:`ClusterSpace` object one has to specify a
    prototype structure and cutoff radii for each cluster order that
    should be included.  For example the following snippet will set up
    a :class:`ClusterSpace` object for a body-centered-cubic (BCC)
    structure including second order terms up to a distance of 5 A and
    third order terms up to a distance of 4 A.

    >>> from ase.build import bulk
    >>> prim = bulk('W')
    >>> cs = ClusterSpace(prim, [5.0, 4.0])
    """
    # TODO: This class probably need some more documentation and a reference to
    # the thesis
    # TODO: Fix doc for n-body cutoff

    def __init__(self, prototype_structure, cutoffs, config=None,
                 cluster_filter=None, **kwargs):

        if not all(prototype_structure.pbc):
            raise ValueError('prototype_structure must have pbc.')

        if isinstance(cutoffs, Cutoffs):
            self._cutoffs = cutoffs
        elif isinstance(cutoffs, list):
            self._cutoffs = CutoffMaximumBody(cutoffs, len(cutoffs) + 1)
        else:
            raise TypeError('cutoffs must be a list or a Cutoffs object')

        if config is None:
            config = Config(**kwargs)
        else:
            if not isinstance(config, Config):
                raise TypeError('config kw must be of type {}'.format(Config))
            if len(kwargs) > 0:
                raise ValueError('use either Config or kwargs, not both')
        self._config = config

        if cluster_filter is None:
            self._cluster_filter = BaseClusterFilter()
        else:
            self._cluster_filter = cluster_filter

        self._atom_list = None
        self._cluster_list = None
        self._symmetry_dataset = None
        self._permutations = None
        self._prim = None
        self._orbits = None

        self._constraint_vectors = None
        # TODO: How to handle the constraint matrices? Should they even be
        # stored?
        self._constraint_matrices = None
        # Is this the best way or should the prim be instantiated separately?

        build_cluster_space(self, prototype_structure)

    # TODO: Should everything here be properties? deepcopy/ref etc.?
    # TODO: Docstrings for properties
    @property
    def n_dofs(self):
        """int : number of free parameters in the model

        If the sum rules are not enforced the number of DOFs is the same as
        the total number of eigentensors in all orbits.
        """
        return self._get_n_dofs()

    @property
    def cutoffs(self):
        """ Cutoffs : cutoffs used for constructing the cluster space """
        return deepcopy(self._cutoffs)

    @property
    def symprec(self):
        """ float : symprec value used when constructing the cluster space """
        return self._config['symprec']

    @property
    def acoustic_sum_rules(self):
        """ bool : True if acoustic sum rules are enforced """
        return self._config['acoustic_sum_rules']

    @property
    def length_scale(self):
        """ float : normalization constant of the force constants """
        return self._config['length_scale']

    @property
    def primitive_structure(self):
        """ ase.Atoms : structure of the lattice """
        return self._prim.copy()

    @property
    def spacegroup(self):
        """ str : space group of the lattice structure obtained from spglib """
        return self._symmetry_dataset['international'] + ' ({})'.format(
            self._symmetry_dataset['number'])

    @property
    def wyckoff_sites(self):
        """ list : wyckoff sites in the primitive cell """
        return self._symmetry_dataset['equivalent_atoms']

    @property
    def rotation_matrices(self):
        """ list(numpy.ndarray) : symmetry elements (`3x3` matrices)
        representing rotations """
        return self._symmetry_dataset['rotations'].copy()

    @property
    def translation_vectors(self):
        """ list(numpy.ndarray) : symmetry elements representing
        translations """
        # TODO: bug incoming!
        return (self._symmetry_dataset['translations'] % 1).copy()

    @property
    def permutations(self):
        """ list(numpy.ndarray) : lookup for permutation references """
        return deepcopy(self._permutations)

    @property
    def atom_list(self):
        """ BiMap : atoms inside the cutoff relative to the of the center cell
        """
        return self._atom_list

    @property
    def cluster_list(self):
        """ BiMap : clusters possible within the cutoff """
        return self._cluster_list

    @property
    def orbits(self):  # TODO: add __getitem__ method
        """ list(Orbit) : orbits associated with the lattice structure. """
        return self._orbits

    @property
    def orbit_data(self):
        """ list(dict) : detailed information for each orbit, e.g., cluster
        radius and atom types.
        """
        data = []
        p = 0
        for orbit_index, orbit in enumerate(self.orbits):
            d = {}
            d['index'] = orbit_index
            d['order'] = orbit.order
            d['radius'] = orbit.radius
            d['maximum_distance'] = orbit.maximum_distance
            d['n_clusters'] = len(orbit.orientation_families)
            d['eigentensors'] = orbit.eigentensors
            d['n_parameters'] = len(d['eigentensors'])

            types, wyckoff_sites = [], []
            for atom_index in self.cluster_list[orbit.prototype_index]:
                atom = self.atom_list[atom_index]
                types.append(self.primitive_structure.numbers[atom.site])
                wyckoff_sites.append(self.wyckoff_sites[atom.site])
            d['prototype_cluster'] = self.cluster_list[orbit.prototype_index]
            d['prototype_atom_types'] = tuple(types)
            d['prototype_wyckoff_sites'] = tuple(wyckoff_sites)

            d['geometrical_order'] = len(set(d['prototype_cluster']))
            d['parameter_indices'] = np.arange(p, p + len(orbit.eigentensors))

            p += len(orbit.eigentensors)
            data.append(d)

        return data

    def get_parameter_indices(self, order):
        """
        Returns a list of the parameter indices associated with the requested
        order.

        Parameters
        ----------
        order : int
            order for which to return the parameter indices

        Returns
        -------
        list(int)
            list of parameter indices associated with the requested order

        Raises
        ------
        ValueError
            if the order is not included in the cluster space
         """
        order = int(order)
        if order not in self.cutoffs.orders:
            raise ValueError('Order must be in {}'.format(self.cutoffs.orders))
        min_param = 0
        max_param = 0
        for orbit in self.orbits:
            if orbit.order < order:
                min_param += len(orbit.eigentensors)
                max_param = min_param
            elif orbit.order == order:
                max_param += len(orbit.eigentensors)
            else:
                break
        rows, cols = self._cvs.nonzero()
        parameters = set()
        for r, c in zip(rows, cols):
            if min_param <= r < max_param:
                parameters.add(c)
        for r, c in zip(rows, cols):
            if c in parameters:
                assert min_param <= r < max_param, 'The internals are broken!'

        return sorted(parameters)

    def get_n_dofs_by_order(self, order):
        """ Returns number of degrees of freedom for the given order.

        Parameters
        ----------
        order : int
            order for which to return the number of dofs

        Returns
        -------
        int
            number of degrees of freedom
        """
        return len(self.get_parameter_indices(order=order))

    def _get_n_dofs(self):
        """ Returns the number of degrees of freedom. """
        return self._cvs.shape[1]

    def _map_parameters(self, parameters):
        """ Maps irreducible parameters to the real parameters associated with
        the eigentensors.
        """
        if len(parameters) != self.n_dofs:
            raise ValueError('Invalid number of parameters, please provide {} '
                             'parameters'.format(self.n_dofs))
        return self._cvs.dot(parameters)

    def print_tables(self):
        """ Prints table data, i.e. information as a function of order and
        n-body for the clusterspace. """

        n_rows = self.cutoffs.max_nbody
        n_cols = self.cutoffs.max_order - 1

        # collect cutoff matrix
        cutoff_matrix = self.cutoffs.cutoff_matrix
        cutoff_matrix = np.vstack(([[None] * n_cols], cutoff_matrix))

        # collect cluster, orbit, eigentensor counts
        cluster_counts = np.zeros((n_rows, n_cols), dtype=int)
        orbit_counts = np.zeros((n_rows, n_cols), dtype=int)
        eigentensor_counts = np.zeros((n_rows, n_cols), dtype=int)
        for orbit in self.orbits:
            proto_cluster = self.cluster_list[orbit.prototype_index]
            order = len(proto_cluster)
            nbody = len(set(proto_cluster))
            cluster_counts[nbody-1, order-2] += len(orbit.orientation_families)
            orbit_counts[nbody-1, order-2] += 1
            eigentensor_counts[nbody-1, order-2] += len(orbit.eigentensors)

        # print
        print('Cutoff Matrix')
        print_table(cutoff_matrix)
        print('\nCluster counts')
        print_table(cluster_counts, sum_=True)
        print('\nOrbit counts')
        print_table(orbit_counts, sum_=True)
        print('\nEigentensor counts')
        print_table(eigentensor_counts, sum_=True)

    def print_orbits(self):
        """ Prints a list of all orbits. """
        orbits = self.orbit_data

        def str_orbit(index, orbit):
            elements = ' '.join(chemical_symbols[n] for n in
                                orbit['prototype_atom_types'])
            fields = OrderedDict([
                ('index',        '{:^3}'.format(index)),
                ('order',        '{:^3}'.format(orbit['order'])),
                ('elements',     '{:^18}'.format(elements)),
                ('radius',       '{:^8.4f}'.format(orbit['radius'])),
                ('prototype',    '{:^18}'
                 .format(str(orbit['prototype_cluster']))),
                ('clusters',     '{:^4}'.format(orbit['n_clusters'])),
                ('parameters',   '{:^3}'.format(len(orbit['eigentensors']))),
                ])

            s = []
            for name, value in fields.items():
                n = max(len(name), len(value))
                if index < 0:
                    s += ['{s:^{n}}'.format(s=name, n=n)]
                else:
                    s += ['{s:^{n}}'.format(s=value, n=n)]
            return ' | '.join(s)

        # table header
        width = max(len(str_orbit(-1, orbits[-1])),
                    len(str_orbit(0, orbits[-1])))
        print(' List of Orbits '.center(width, '='))
        print(str_orbit(-1, orbits[0]))
        print(''.center(width, '-'))

        # table body
        for i, orbit in enumerate(orbits):
            print(str_orbit(i, orbit))
        print(''.center(width, '='))

    def __str__(self):

        def str_order(order, header=False):
            formats = {'order':        '{:2}',
                       'num_orbits':   '{:5}',
                       'num_params':   '{:5}',
                       'num_clusters': '{:5}'}
            s = []
            for name, value in order.items():
                str_repr = formats[name].format(value)
                n = max(len(name), len(str_repr))
                if header:
                    s += ['{s:^{n}}'.format(s=name, n=n)]
                else:
                    s += ['{s:^{n}}'.format(s=str_repr, n=n)]
            return ' | '.join(s)

        # collect data
        orbits = self.orbit_data
        orders = OrderedDict(
            (order, OrderedDict(order=order)) for order in self.cutoffs.orders)
        for orbit in orbits:
            k = orbit['order']
            orders[k]['num_orbits'] = orders[k].get('num_orbits', 0) + 1
            orders[k]['num_clusters'] = orders[k].get('num_clusters', 0) + \
                orbit['n_clusters']

        # basic information
        prototype = orders[list(orders.keys())[-1]]
        n = max(len(str_order(prototype)), 54)
        s = []
        s.append(' Cluster Space '.center(n, '='))
        data = [('Spacegroup',               self.spacegroup),
                ('symprec',                  self.symprec),
                ('Sum rules',                self.acoustic_sum_rules),
                ('Length scale',             self.length_scale),
                ('Cutoffs',                  self.cutoffs),
                ('Cell',                     self.primitive_structure.cell),
                ('Basis',                    self.primitive_structure.basis),
                ('Numbers',                  self.primitive_structure.numbers),
                ('Total number of orbits',   len(orbits)),
                ('Total number of clusters',
                 sum([order['num_clusters'] for order in orders.values()])),
                ('Total number of parameters',
                 self._get_n_dofs()
                 )]
        for field, value in data:
            if str(value).count('\n') > 1:
                s.append('{:26} :\n{}'.format(field, value))
            else:
                s.append('{:26} : {}'.format(field, value))

        # table header
        s.append(''.center(n, '-'))
        s.append(str_order(prototype, header=True))
        s.append(''.center(n, '-'))
        for order in orders.values():
            s.append(str_order(order).rstrip())
        s.append(''.center(n, '='))
        return '\n'.join(s)

    def __repr__(self):
        s = 'ClusterSpace({!r}, {!r}, {!r})'
        return s.format(self.primitive_structure, self.cutoffs, self._config)

    def copy(self):
        return deepcopy(self)

    def write(self, fileobj):
        """ Writes cluster space to file.

        The instance is saved into a custom format based on tar-files. The
        resulting file will be a valid tar file and can be browsed by by a tar
        reader. The included objects are themself either pickles, npz or other
        tars.

        Parameters
        ----------
        fileobj : str or file-like object
            If the input is a string a tar archive will be created in the
            current directory. Otherwise the input must be a valid file
            like object.
        """
        # Create a tar archive
        if isinstance(fileobj, str):
            tar_file = tarfile.open(name=fileobj, mode='w')
        else:
            tar_file = tarfile.open(fileobj=fileobj, mode='w')

        # Attributes in pickle format
        pickle_attributes = ['_config',
                             '_symmetry_dataset', '_permutations',
                             '_atom_list', '_cluster_list']
        items_pickle = dict()
        for attribute in pickle_attributes:
            items_pickle[attribute] = self.__getattribute__(attribute)
        add_items_to_tarfile_pickle(tar_file, items_pickle, 'attributes')

        # Constraint matrices and vectors in hdf5 format
        items = dict(cvs=self._cvs)
        add_items_to_tarfile_pickle(tar_file, items, 'constraint_vectors')

        # Cutoffs and prim with their builtin write/read functions
        items_custom = {'_cutoffs': self._cutoffs, '_prim': self._prim}
        add_items_to_tarfile_custom(tar_file, items_custom)

        # Orbits
        add_list_to_tarfile_custom(tar_file, self._orbits, 'orbits')
        add_list_to_tarfile_custom(tar_file, self._dropped_orbits,
                                   'dropped_orbits')

        # Done!
        tar_file.close()

    def read(f):
        """ Reads a cluster space from file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to load from (file object)
        """

        # Instantiate empty cs obj.
        cs = ClusterSpace.__new__(ClusterSpace)

        # Load from file on disk or file-like
        if type(f) is str:
            tar_file = tarfile.open(mode='r', name=f)
        else:
            tar_file = tarfile.open(mode='r', fileobj=f)

        # Attributes
        attributes = read_items_pickle(tar_file, 'attributes')
        for name, value in attributes.items():
            cs.__setattr__(name, value)

        # Load the constraint matrices into their dict
        items = read_items_pickle(tar_file, 'constraint_vectors')
        cs._cvs = items['cvs']

        # Cutoffs and prim via custom save funcs
        fileobj = tar_file.extractfile('_cutoffs')
        cs._cutoffs = Cutoffs.read(fileobj)

        fileobj = tar_file.extractfile('_prim')
        cs._prim = Atoms.read(fileobj)

        # Orbits are stored in a separate archive
        cs._orbits = read_list_custom(tar_file, 'orbits', Orbit.read)
        cs._dropped_orbits = read_list_custom(
            tar_file, 'dropped_orbits', Orbit.read)

        tar_file.close()
        return cs



#!/usr/bin/env python3

import re
import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 5, 0, 'final', 0):
    raise SystemExit('Python 3.5 or later is required!')

with open('README.rst', encoding='utf-8') as fd:
    long_description = fd.read()

with open('hiphive/__init__.py') as fd:
    lines = '\n'.join(fd.readlines())

version = re.search("__version__ = '(.*)'", lines).group(1)
maintainer = re.search("__maintainer__ = '(.*)'", lines).group(1)
maintainer_email = re.search("__maintainer_email__ = '(.*)'", lines).group(1)
url = re.search("__url__ = '(.*)'", lines).group(1)
license = re.search("__license__ = '(.*)'", lines).group(1)
description = re.search("__description__ = '(.*)'", lines).group(1)

# PyPI name
name = 'hiphive'

setup(name=name,
      version=version,
      description=description,
      long_description=long_description,
      url=url,
      maintainer=maintainer,
      maintainer_email=maintainer_email,
      platforms=['unix'],
      install_requires=['ase',
                        'h5py',
                        'numba',
                        'numpy>=1.12',
                        'sklearn',
                        'spglib',
                        'sympy'],
      packages=find_packages(),
      classifiers=[
          'Development Status :: 4 - Beta',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Topic :: Scientific/Engineering :: Physics'])



import numba
import numpy as np


@numba.jit(nopython=True)
def _get_forces(clusters, force_constants,
                atom_indices_list, atom_positions_list, atom_counts_list,
                prefactors_list,
                nbody, order,
                forces, E,
                displacements):
    fc_tmp = np.zeros(3**order)
    fc = np.zeros(3**order)
    forces_tmp = np.zeros(forces.shape)
    f = np.zeros(3)
    for i in range(len(clusters)):
        cluster = clusters[i]
        fc[:] = force_constants[i].flat
        atom_positions = atom_positions_list[i]
        prefactors = prefactors_list[i]
        cluster_force_contribution(atom_positions, prefactors, nbody,
                                   fc_tmp, fc, order,
                                   displacements,
                                   cluster, f, forces_tmp)
    forces += forces_tmp
    E_tmp = 0
    for atom_index in range(len(forces_tmp)):
        E_tmp += np.dot(forces_tmp[atom_index], displacements[atom_index])
    E -= E_tmp / order


def cluster_force_contribution_einsum(positions, prefactors, numbers,
                                      fc_tmp, fc, order,
                                      disps,
                                      cluster, f, F):
    fc = fc.reshape((3,)*order)
    for p, pf in zip(positions, prefactors):
        t = [fc, list(range(order))]
        for i, a in enumerate(cluster):
            if i == p:
                continue
            t.append(disps[a])
            t.append([i])
        F[cluster[p], :] += pf * np.einsum(*t)


try:
    import numba

    @numba.jit(nopython=True)
    def tvl(fc, v, order):
        """Contracts a 3-dim tensor of rank n with a 3-dim vector from the left
        """
        t = 3**(order-1)
        fc[0:t] *= v[0]
        fc[t:2*t] *= v[1]
        fc[2*t:3*t] *= v[2]
        fc[0:t] = fc[0:t] + fc[t:2*t] + fc[2*t:3*t]

    @numba.jit(nopython=True)
    def tvr(fc, v, order):
        for i in range(3**(order-1)):
            t = 3*i
            fc[i] = fc[t]*v[0]+fc[t+1]*v[1]+fc[t+2]*v[2]

    @numba.jit(nopython=True)
    def contraction(fc, d, cluster, pos, order, prefac, f):
        i = 0
        for l in range(pos):
            v = d[cluster[l], :]
            tvl(fc, v, order - i)
            i += 1
        for r in range(order - pos - 1):
            v = d[cluster[order - r - 1], :]
            tvr(fc, v, order - i)
            i += 1
        f[0] = fc[0] * prefac
        f[1] = fc[1] * prefac
        f[2] = fc[2] * prefac

    @numba.jit(nopython=True)
    def cluster_force_contribution_numba(positions, prefactors, numbers,
                                         fc_tmp, fc, order,
                                         disps,
                                         cluster, f, F):
        for i in range(numbers):
            pos = positions[i]
            prefac = prefactors[i]
            for j in range(3**order):
                fc_tmp[j] = fc[j]
            contraction(fc_tmp, disps,
                        cluster, pos, order, prefac, f)
            a = cluster[pos]
            F[a, 0] += f[0]
            F[a, 1] += f[1]
            F[a, 2] += f[2]

    @numba.jit(nopython=True)
    def clusters_force_contribution_numba(positions, prefactors, numbers,
                                          fc_tmp, fc, order,
                                          disps,
                                          clusters, f, F):
        for i in range(clusters.shape[0]):
            cluster = clusters[i, :]
            cluster_force_contribution_numba(positions, prefactors, numbers,
                                             fc_tmp, fc, order,
                                             disps,
                                             cluster, f, F)

    cluster_force_contribution = cluster_force_contribution_numba
    clusters_force_contribution = clusters_force_contribution_numba

except ImportError:

    cluster_force_contribution = cluster_force_contribution_einsum



"""
Collection of tools for managing and analyzing linear models.

Todo
-----
Consider what functionality we actually want here

"""

import numpy as np
from collections import namedtuple


ScatterData = namedtuple('ScatterData', ['target', 'predicted'])


def compute_correlation_matrix(A: np.ndarray) -> np.ndarray:
    """
    Returns the correlation matrix for the rows in the fit matrix.

    Notes
    -----
    Naive implementation.

    Parameters
    ----------
    A
        fit matrix
    """
    N = A.shape[0]
    C = np.zeros((N, N))
    for i in range(N):
        for j in range(i+1, N):
            norm = np.linalg.norm(A[i, :]) * np.linalg.norm(A[j, :])
            c_ij = np.dot(A[i, :], A[j, :]) / norm
            C[i, j] = c_ij
            C[j, i] = c_ij
    return C
